#
# completion.rb
#
#   Copyright (c) 2001,2002 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

unless Array.instance_methods(false).include?('select')
class Array
  def select( *args )
    if block_given?
      super
    else
      self.indexes(*args)
    end
  end
end
end

unless Enumerable.instance_methods(false).include?('reject')
module Enumerable
  def reject
    list = []
    each do |i|
      list.push i unless yield(i)
    end
    list
  end
end
end

module Enumerable
  def select_by( method )
    ret = []
    __send__(method) {|a| yield(a) and ret.push a }
    ret
  end
end


module ReFe

  class CompletionTable

    def initialize( cand = nil )
      @candidates = {}
      @table = {}
      if cand
        cand.each do |c|
          add c
        end
      end
    end

    def inspect
      "#<#{self.class}>"
    end

    def add( cand )
      return if @candidates[cand]
      @candidates[cand] = true

      str = cand.downcase
      begin
        (@table[str] ||= []).push cand
        str.chop!
      end until str.empty?
    end

    def candidates
      @candidates.keys
    end

    def candidate?( name )
      @candidates.key? name
    end

    def complete( name )
      if @candidates[name]
        [name]
      else
        (@table[name.downcase] || []).dup
      end
    end

    def complete_method( name )
      if @candidates[name]
        return [name]
      end
      case name[-1,1]
      when /[?!=]/
        tail = name[-1]
        (@table[name.chop.downcase] || []).select {|m| m[-1] == tail }
      when /-/
        tail = name[-1]
        (@table[name.chop.downcase] || []).reject {|m| /[?!=]\z/ === m }
      else
        (@table[name.downcase] || []).dup
      end
    end
  
  end   # class CompletionTable


  class FCompletionTable

    def initialize( cand = nil )
      @tree = {}
      if cand
        cand.each do |c|
          add c
        end
      end
    end

    def inspect
      "#<#{self.class}>"
    end

    def add( cand )
      a = cand.split(/_/)
      ident = a.pop
      t = @tree
      a.each do |prefix|
        t = (t[prefix] ||= {})
      end
      (t['-ident'] ||= {})[ ident ] = true
    end

    def complete( name )
      _comp(@tree, name.split(/_/), '#').map {|n| n.sub(/\A\#_/, '') }
    end

    private

    def _comp( table, keys, path )
      k = keys[0]
      keys = keys[1..-1]
      k or return collect_all_sub_candidates(table, path)

      if complete_match_at_tail? table, keys, k
        return [path + '_' + k]
      end
      if complete_match? table, k
        ret = _comp( table[k], keys, path + '_' + k )
        return ret unless ret.empty?
      end

      find_candidates_at_tail(      table, keys, k, path) +
      find_candidates_from_children(table, keys, k, path)
    end

    def complete_match_at_tail?( table, keys, k )
      keys.empty? and it = table['-ident'] and it[k]
    end

    def complete_match?( table, k )
      table.key? k
    end

    def find_candidates_at_tail( table, keys, k, path )
      if keys.empty? and it = table['-ident']
        complete_word(it, k).map {|n| path + '_' + n }
      else
        []
      end
    end

    def find_candidates_from_children( table, keys, k, path )
      complete_word(table, k).map {|n|
              _comp table[n], keys, path + '_' + n }.flatten
    end

    def complete_word( table, key )
      cands = []

      if /[A-Z]/ === key
        table.select_by(:each_key) {|k| k.index(key) == 0 }
      else
        table.select_by(:each_key) {|k| k.downcase.index(key) == 0 }
      end
    end

    def collect_all_sub_candidates( table, path )
      collect_recursive(table).map {|n| path + '_' + n }
    end

    def collect_recursive( table )
      table.map {|k,t|
          if k == '-ident'
            t.keys
          else
            collect_recursive(t).map {|n| k + '_' + n }
          end
      }.flatten
    end
  
  end

end
